@@ -3,12 +3,13 @@ import asyncio.events
 import functools
 import inspect
 import io
+import numbers
 import os
 import re
-import sys
 import threading
 from contextlib import contextmanager
 from glob import has_magic
+from typing import Iterable
 
 from .callbacks import _DEFAULT_CALLBACK
 from .exceptions import FSTimeoutError
@@ -19,12 +20,13 @@ private = re.compile("_[^_]")
 iothread = [None]  # dedicated fsspec IO thread
 loop = [None]  # global event loop for any non-async instance
 _lock = None  # global lock placeholder
+get_running_loop = asyncio.get_running_loop
 
 
 def get_lock():
     """Allocate or return a threading lock.
 
-    The lock is allocatted on first use to allow setting one lock per forked process.
+    The lock is allocated on first use to allow setting one lock per forked process.
     """
     global _lock
     if not _lock:
@@ -61,8 +63,10 @@ def sync(loop, func, *args, timeout=None, **kwargs):
     """
     Make loop run coroutine until it returns. Runs in other thread
 
-    Example usage:
-        fsspec.asyn.sync(fsspec.asyn.get_loop(), func, *args, timeout=timeout, **kwargs)
+    Examples
+    --------
+    >>> fsspec.asyn.sync(fsspec.asyn.get_loop(), func, *args,
+                         timeout=timeout, **kwargs)
     """
     timeout = timeout if timeout else None  # convert 0 or 0.0 to None
     # NB: if the loop is not running *yet*, it is OK to submit work
@@ -117,11 +121,7 @@ def sync_wrapper(func, obj=None):
 def _selector_policy():
     original_policy = asyncio.get_event_loop_policy()
     try:
-        if (
-            sys.version_info >= (3, 8)
-            and os.name == "nt"
-            and hasattr(asyncio, "WindowsSelectorEventLoopPolicy")
-        ):
+        if os.name == "nt" and hasattr(asyncio, "WindowsSelectorEventLoopPolicy"):
             asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
 
         yield
@@ -129,17 +129,6 @@ def _selector_policy():
         asyncio.set_event_loop_policy(original_policy)
 
 
-def get_running_loop():
-    if hasattr(asyncio, "get_running_loop"):
-        return asyncio.get_running_loop()
-    else:
-        loop = asyncio._get_running_loop()
-        if loop is None:
-            raise RuntimeError("no running event loop")
-        else:
-            return loop
-
-
 def get_loop():
     """Create or return the default fsspec IO loop
 
@@ -159,25 +148,6 @@ def get_loop():
     return loop[0]
 
 
-@contextmanager
-def fsspec_loop():
-    """Temporarily switch the current event loop to the fsspec's
-    own loop, and then revert it back after the context gets
-    terminated.
-    """
-    try:
-        original_loop = get_running_loop()
-    except RuntimeError:
-        original_loop = None
-
-    fsspec_loop = get_loop()
-    try:
-        asyncio._set_running_loop(fsspec_loop)
-        yield fsspec_loop
-    finally:
-        asyncio._set_running_loop(original_loop)
-
-
 try:
     import resource
 except ImportError:
@@ -215,6 +185,15 @@ def _get_batch_size(nofiles=False):
         return soft_limit // 8
 
 
+def running_async() -> bool:
+    """Being executed by an event loop?"""
+    try:
+        asyncio.get_running_loop()
+        return True
+    except RuntimeError:
+        return False
+
+
 async def _run_coros_in_chunks(
     coros,
     batch_size=None,
@@ -307,6 +286,7 @@ class AsyncFileSystem(AbstractFileSystem):
     # for _* methods and inferred for overridden methods.
 
     async_impl = True
+    mirror_sync_methods = True
     disable_throttling = False
 
     def __init__(self, *args, asynchronous=False, loop=None, batch_size=None, **kwargs):
@@ -357,7 +337,15 @@ class AsyncFileSystem(AbstractFileSystem):
             on_error = "raise"
 
         paths = await self._expand_path(path1, maxdepth=maxdepth, recursive=recursive)
-        path2 = other_paths(paths, path2)
+        isdir = isinstance(path2, str) and (
+            path2.endswith("/") or await self._isdir(path2)
+        )
+        path2 = other_paths(
+            paths,
+            path2,
+            exists=isdir and isinstance(path1, str) and not path1.endswith("/"),
+            is_dir=isdir,
+        )
         batch_size = batch_size or self.batch_size
         coros = [self._cp_file(p1, p2, **kwargs) for p1, p2 in zip(paths, path2)]
         result = await _run_coros_in_chunks(
@@ -369,6 +357,9 @@ class AsyncFileSystem(AbstractFileSystem):
                 continue
             raise ex
 
+    async def _pipe_file(self, path, value, **kwargs):
+        raise NotImplementedError
+
     async def _pipe(self, path, value=None, batch_size=None, **kwargs):
         if isinstance(path, str):
             path = {path: value}
@@ -401,7 +392,7 @@ class AsyncFileSystem(AbstractFileSystem):
                     end = size + end
             elif end is None:
                 end = ""
-            if isinstance(end, int):
+            if isinstance(end, numbers.Integral):
                 end -= 1  # bytes range is inclusive
         return "bytes=%s-%s" % (start, end)
 
@@ -435,7 +426,14 @@ class AsyncFileSystem(AbstractFileSystem):
             return out[0]
 
     async def _cat_ranges(
-        self, paths, starts, ends, max_gap=None, batch_size=None, **kwargs
+        self,
+        paths,
+        starts,
+        ends,
+        max_gap=None,
+        batch_size=None,
+        on_error="return",
+        **kwargs,
     ):
         # TODO: on_error
         if max_gap is not None:
@@ -443,9 +441,9 @@ class AsyncFileSystem(AbstractFileSystem):
             raise NotImplementedError
         if not isinstance(paths, list):
             raise TypeError
-        if not isinstance(starts, list):
+        if not isinstance(starts, Iterable):
             starts = [starts] * len(paths)
-        if not isinstance(ends, list):
+        if not isinstance(ends, Iterable):
             ends = [starts] * len(paths)
         if len(starts) != len(paths) or len(ends) != len(paths):
             raise ValueError
@@ -454,7 +452,9 @@ class AsyncFileSystem(AbstractFileSystem):
             for p, s, e in zip(paths, starts, ends)
         ]
         batch_size = batch_size or self.batch_size
-        return await _run_coros_in_chunks(coros, batch_size=batch_size, nofiles=True)
+        return await _run_coros_in_chunks(
+            coros, batch_size=batch_size, nofiles=True, return_exceptions=True
+        )
 
     async def _put_file(self, lpath, rpath, **kwargs):
         raise NotImplementedError
@@ -488,8 +488,14 @@ class AsyncFileSystem(AbstractFileSystem):
             lpath = make_path_posix(lpath)
         fs = LocalFileSystem()
         lpaths = fs.expand_path(lpath, recursive=recursive)
+        isdir = isinstance(rpath, str) and (
+            rpath.endswith("/") or await self._isdir(rpath)
+        )
         rpaths = other_paths(
-            lpaths, rpath, exists=isinstance(rpath, str) and await self._isdir(rpath)
+            lpaths,
+            rpath,
+            exists=isdir and isinstance(lpath, str) and not lpath.endswith("/"),
+            is_dir=isdir,
         )
 
         is_dir = {l: os.path.isdir(l) for l in lpaths}
@@ -529,12 +535,22 @@ class AsyncFileSystem(AbstractFileSystem):
         constructor, or for all instances by setting the "gather_batch_size" key
         in ``fsspec.config.conf``, falling back to 1/8th of the system limit .
         """
-        from fsspec.implementations.local import make_path_posix
+        from fsspec.implementations.local import LocalFileSystem, make_path_posix
 
+        # First check for rpath trailing slash as _strip_protocol removes it.
+        rpath_trailing_slash = isinstance(rpath, str) and rpath.endswith("/")
         rpath = self._strip_protocol(rpath)
         lpath = make_path_posix(lpath)
         rpaths = await self._expand_path(rpath, recursive=recursive)
-        lpaths = other_paths(rpaths, lpath)
+        isdir = isinstance(lpath, str) and (
+            lpath.endswith("/") or LocalFileSystem().isdir(lpath)
+        )
+        lpaths = other_paths(
+            rpaths,
+            lpath,
+            exists=isdir and not rpath_trailing_slash,
+            is_dir=isdir,
+        )
         [os.makedirs(os.path.dirname(lp), exist_ok=True) for lp in lpaths]
         batch_size = kwargs.pop("batch_size", self.batch_size)
 
@@ -582,6 +598,9 @@ class AsyncFileSystem(AbstractFileSystem):
         raise NotImplementedError
 
     async def _walk(self, path, maxdepth=None, **kwargs):
+        if maxdepth is not None and maxdepth < 1:
+            raise ValueError("maxdepth must be at least 1")
+
         path = self._strip_protocol(path)
         full_dirs = {}
         dirs = {}
@@ -732,11 +751,12 @@ class AsyncFileSystem(AbstractFileSystem):
             return {name: out[name] for name in names}
 
     async def _expand_path(self, path, recursive=False, maxdepth=None):
+        if maxdepth is not None and maxdepth < 1:
+            raise ValueError("maxdepth must be at least 1")
+
         if isinstance(path, str):
             out = await self._expand_path([path], recursive, maxdepth)
         else:
-            # reduce depth on each recursion level unless None or 0
-            maxdepth = maxdepth if not maxdepth else maxdepth - 1
             out = set()
             path = [self._strip_protocol(p) for p in path]
             for p in path:  # can gather here
@@ -756,6 +776,8 @@ class AsyncFileSystem(AbstractFileSystem):
                 if p not in out and (recursive is False or (await self._exists(p))):
                     # should only check once, for the root
                     out.add(p)
+            # reduce depth on each recursion level unless None or 0
+            maxdepth = maxdepth if not maxdepth else maxdepth - 1
         if not out:
             raise FileNotFoundError(path)
         return list(sorted(out))
