@@ -22,7 +22,6 @@ from .utils import (
     build_name_function,
     infer_compression,
     stringify_path,
-    update_storage_options,
 )
 
 logger = logging.getLogger("fsspec")
@@ -36,7 +35,7 @@ class OpenFile:
     are typically binary-only.
 
     These instances are safe to serialize, as the low-level file object
-    is not created until invoked using `with`.
+    is not created until invoked using ``with``.
 
     Parameters
     ----------
@@ -121,10 +120,6 @@ class OpenFile:
     def __exit__(self, *args):
         self.close()
 
-    def __del__(self):
-        if hasattr(self, "fobjects"):
-            self.close()
-
     @property
     def full_name(self):
         return _unstrip_protocol(self.path, self.fs)
@@ -305,35 +300,23 @@ def open_files(
 
 
 def _un_chain(path, kwargs):
-    if isinstance(path, (tuple, list)):
-        bits = [_un_chain(p, kwargs) for p in path]
-        out = []
-        for pbit in zip(*bits):
-            paths, protocols, kwargs = zip(*pbit)
-            if len(set(protocols)) > 1:
-                raise ValueError("Protocol mismatch in URL chain")
-            if len(set(paths)) == 1:
-                paths = paths[0]
-            else:
-                paths = list(paths)
-            out.append([paths, protocols[0], kwargs[0]])
-        return out
     x = re.compile(".*[^a-z]+.*")  # test for non protocol-like single word
     bits = (
         [p if "://" in p or x.match(p) else p + "://" for p in path.split("::")]
         if "::" in path
         else [path]
     )
-    if len(bits) < 2:
-        return []
     # [[url, protocol, kwargs], ...]
     out = []
     previous_bit = None
+    kwargs = kwargs.copy()
     for bit in reversed(bits):
-        protocol = split_protocol(bit)[0] or "file"
+        protocol = kwargs.pop("protocol", None) or split_protocol(bit)[0] or "file"
         cls = get_filesystem_class(protocol)
         extra_kwargs = cls._get_kwargs_from_urls(bit)
-        kws = kwargs.get(protocol, {})
+        kws = kwargs.pop(protocol, {})
+        if bit is bits[0]:
+            kws.update(kwargs)
         kw = dict(**extra_kwargs, **kws)
         bit = cls._strip_protocol(bit)
         if (
@@ -368,27 +351,18 @@ def url_to_fs(url, **kwargs):
         The file-systems-specific URL for ``url``.
     """
     chain = _un_chain(url, kwargs)
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = protocol
-            inkwargs["fo"] = urls
-        urlpath, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-    else:
-        protocol = split_protocol(url)[0]
-        cls = get_filesystem_class(protocol)
-
-        options = cls._get_kwargs_from_urls(url)
-        update_storage_options(options, kwargs)
-        fs = cls(**options)
-        urlpath = fs._strip_protocol(url)
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = protocol
+        inkwargs["fo"] = urls
+    urlpath, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
     return fs, urlpath
 
 
@@ -591,57 +565,34 @@ def get_fs_token_paths(
     if isinstance(urlpath, (list, tuple, set)):
         if not urlpath:
             raise ValueError("empty urlpath sequence")
-        urlpath = [stringify_path(u) for u in urlpath]
+        urlpath0 = stringify_path(list(urlpath)[0])
     else:
-        urlpath = stringify_path(urlpath)
-    chain = _un_chain(urlpath, storage_options or {})
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, nested_protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = nested_protocol
-            inkwargs["fo"] = urls
-        paths, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-        if isinstance(paths, (list, tuple, set)):
-            paths = [fs._strip_protocol(u) for u in paths]
-        else:
-            paths = fs._strip_protocol(paths)
+        urlpath0 = stringify_path(urlpath)
+    storage_options = storage_options or {}
+    if protocol:
+        storage_options["protocol"] = protocol
+    chain = _un_chain(urlpath0, storage_options or {})
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, nested_protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = nested_protocol
+        inkwargs["fo"] = urls
+    paths, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
+    if isinstance(urlpath, (list, tuple, set)):
+        pchains = [
+            _un_chain(stringify_path(u), storage_options or {})[0] for u in urlpath
+        ]
+        if len(set(pc[1] for pc in pchains)) > 1:
+            raise ValueError("Protocol mismatch getting fs from %s", urlpath)
+        paths = [pc[0] for pc in pchains]
     else:
-        if isinstance(urlpath, (list, tuple, set)):
-            protocols, paths = zip(*map(split_protocol, urlpath))
-            if protocol is None:
-                protocol = protocols[0]
-                if not all(p == protocol for p in protocols):
-                    raise ValueError(
-                        "When specifying a list of paths, all paths must "
-                        "share the same protocol"
-                    )
-            cls = get_filesystem_class(protocol)
-            optionss = list(map(cls._get_kwargs_from_urls, urlpath))
-            paths = [cls._strip_protocol(u) for u in urlpath]
-            options = optionss[0]
-            if not all(o == options for o in optionss):
-                raise ValueError(
-                    "When specifying a list of paths, all paths must "
-                    "share the same file-system options"
-                )
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-        else:
-            protocols = split_protocol(urlpath)[0]
-            protocol = protocol or protocols
-            cls = get_filesystem_class(protocol)
-            options = cls._get_kwargs_from_urls(urlpath)
-            paths = cls._strip_protocol(urlpath)
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-
+        paths = fs._strip_protocol(paths)
     if isinstance(paths, (list, tuple, set)):
         paths = expand_paths_if_needed(paths, mode, num, fs, name_function)
     else:
