@@ -1,3 +1,4 @@
+import ast
 import base64
 import io
 import itertools
@@ -12,15 +13,25 @@ try:
 except ImportError:
     import json
 
-from ..asyn import AsyncFileSystem, sync
+from ..asyn import AsyncFileSystem
 from ..callbacks import _DEFAULT_CALLBACK
 from ..core import filesystem, open, split_protocol
 from ..spec import AbstractFileSystem
-from ..utils import isfilelike, merge_offset_ranges
+from ..utils import isfilelike, merge_offset_ranges, other_paths
 
 logger = logging.getLogger("fsspec.reference")
 
 
+class ReferenceNotReachable(RuntimeError):
+    def __init__(self, reference, target, *args):
+        super().__init__(*args)
+        self.reference = reference
+        self.target = target
+
+    def __str__(self):
+        return f'Reference "{self.reference}" failed to fetch target {self.target}'
+
+
 def _first(d):
     return list(d.values())[0]
 
@@ -78,7 +89,6 @@ class ReferenceFileSystem(AsyncFileSystem):
         simple_templates=True,
         max_gap=64_000,
         max_block=256_000_000,
-        loop=None,
         **kwargs,
     ):
         """
@@ -125,24 +135,25 @@ class ReferenceFileSystem(AsyncFileSystem):
         max_gap, max_block: int
             For merging multiple concurrent requests to the same remote file.
             Neighboring byte ranges will only be merged when their
-            inter-range gap is <= `max_gap`. Default is 64KB. Set to 0
+            inter-range gap is <= ``max_gap``. Default is 64KB. Set to 0
             to only merge when it requires no extra bytes. Pass a negative
             number to disable merging, appropriate for local target files.
             Neighboring byte ranges will only be merged when the size of
-            the aggregated range is <= `max_block`. Default is 256MB.
+            the aggregated range is <= ``max_block``. Default is 256MB.
         kwargs : passed to parent class
         """
-        super().__init__(loop=loop, **kwargs)
+        super().__init__(**kwargs)
         self.target = target
-        self.dataframe = False
         self.template_overrides = template_overrides
         self.simple_templates = simple_templates
         self.templates = {}
         self.fss = {}
+        self._dircache = {}
         self.max_gap = max_gap
         self.max_block = max_block
         if hasattr(fo, "read"):
-            text = fo.read()
+            text = json.load(fo)
+            text = text.decode() if isinstance(text, bytes) else text
         elif isinstance(fo, str):
             if target_protocol:
                 extra = {"protocol": target_protocol}
@@ -152,14 +163,11 @@ class ReferenceFileSystem(AsyncFileSystem):
             # text JSON
             with open(fo, "rb", **dic) as f:
                 logger.info("Read reference from URL %s", fo)
-                text = f.read()
+                text = json.load(f)
         else:
             # dictionaries
             text = fo
-        if self.dataframe:
-            self._process_dataframe()
-        else:
-            self._process_references(text, template_overrides)
+        self._process_references(text, template_overrides)
         if isinstance(fs, dict):
             self.fss = {
                 k: (
@@ -186,7 +194,7 @@ class ReferenceFileSystem(AsyncFileSystem):
                     ref = ref()
                 protocol, _ = fsspec.core.split_protocol(ref)
                 if protocol and protocol not in self.fss:
-                    fs = filesystem(protocol, loop=loop, **(remote_options or {}))
+                    fs = filesystem(protocol, **(remote_options or {}))
                     self.fss[protocol] = fs
         if remote_protocol is None:
             # get single protocol from references
@@ -196,24 +204,22 @@ class ReferenceFileSystem(AsyncFileSystem):
                 if isinstance(ref, list) and ref[0]:
                     protocol, _ = fsspec.core.split_protocol(ref[0])
                     if protocol and protocol not in self.fss:
-                        fs = filesystem(protocol, loop=loop, **(remote_options or {}))
+                        fs = filesystem(protocol, **(remote_options or {}))
                         self.fss[protocol] = fs
 
         if remote_protocol and remote_protocol not in self.fss:
-            fs = filesystem(remote_protocol, loop=loop, **(remote_options or {}))
+            fs = filesystem(remote_protocol, **(remote_options or {}))
             self.fss[remote_protocol] = fs
 
         self.fss[None] = fs or filesystem("file")  # default one
 
-    @property
-    def loop(self):
-        inloop = [fs.loop for fs in self.fss.values() if fs.async_impl]
-        return inloop[0] if inloop else self._loop
-
     def _cat_common(self, path, start=None, end=None):
         path = self._strip_protocol(path)
         logger.debug(f"cat: {path}")
-        part = self.references[path]
+        try:
+            part = self.references[path]
+        except KeyError:
+            raise FileNotFoundError(path)
         if isinstance(part, str):
             part = part.encode()
         if isinstance(part, bytes):
@@ -254,15 +260,20 @@ class ReferenceFileSystem(AsyncFileSystem):
         if isinstance(part_or_url, bytes):
             return part_or_url[start:end]
         protocol, _ = split_protocol(part_or_url)
-        return await self.fss[protocol]._cat_file(part_or_url, start=start, end=end)
+        try:
+            await self.fss[protocol]._cat_file(part_or_url, start=start, end=end)
+        except Exception as e:
+            raise ReferenceNotReachable(path, part_or_url) from e
 
     def cat_file(self, path, start=None, end=None, **kwargs):
         part_or_url, start0, end0 = self._cat_common(path, start=start, end=end)
         if isinstance(part_or_url, bytes):
             return part_or_url[start:end]
         protocol, _ = split_protocol(part_or_url)
-        # TODO: start and end should be passed to cat_file, not sliced
-        return self.fss[protocol].cat_file(part_or_url, start=start0, end=end0)
+        try:
+            return self.fss[protocol].cat_file(part_or_url, start=start0, end=end0)
+        except Exception as e:
+            raise ReferenceNotReachable(path, part_or_url) from e
 
     def pipe_file(self, path, value, **_):
         """Temporarily add binary data or reference as a file"""
@@ -288,18 +299,19 @@ class ReferenceFileSystem(AsyncFileSystem):
         callback.absolute_update(len(data))
 
     def get(self, rpath, lpath, recursive=False, **kwargs):
-        if isinstance(lpath, list):
-            # because we have to figure out here which lpath goes with which path
-            # after grouping
-            raise NotImplementedError
-        proto_dict = _protocol_groups(rpath, self.references)
-        for proto, paths in proto_dict.items():
-            if self.fss[proto].async_impl:
-                sync(self.loop, self._get, paths, lpath, recursive, **kwargs)
-            else:
-                AbstractFileSystem.get(
-                    self, paths, lpath, recursive=recursive, **kwargs
-                )
+        if recursive:
+            # trigger directory build
+            self.ls("")
+        rpath = self.expand_path(rpath, recursive=recursive)
+        fs = fsspec.filesystem("file", auto_mkdir=True)
+        targets = other_paths(rpath, lpath)
+        if recursive:
+            data = self.cat([r for r in rpath if not self.isdir(r)])
+        else:
+            data = self.cat(rpath)
+        for remote, local in zip(rpath, targets):
+            if remote in data:
+                fs.pipe_file(local, data[remote])
 
     def cat(self, path, recursive=False, on_error="raise", **kwargs):
         if isinstance(path, str) and recursive:
@@ -310,7 +322,22 @@ class ReferenceFileSystem(AsyncFileSystem):
         out = {}
         for proto, paths in proto_dict.items():
             fs = self.fss[proto]
-            urls, starts, ends = zip(*[self._cat_common(p) for p in paths])
+            urls, starts, ends = [], [], []
+            for p in paths:
+                # find references or label not-found. Early exit if any not
+                # found and on_error is "raise"
+                try:
+                    u, s, e = self._cat_common(p)
+                    urls.append(u)
+                    starts.append(s)
+                    ends.append(e)
+                except FileNotFoundError as e:
+                    if on_error == "raise":
+                        raise
+                    if on_error != "omit":
+                        out[p] = e
+
+            # process references into form for merging
             urls2 = []
             starts2 = []
             ends2 = []
@@ -329,66 +356,56 @@ class ReferenceFileSystem(AsyncFileSystem):
                     ends2.append(e)
                     paths2.append(p)
             for u, s, e, p in zip(urls, starts, ends, paths):
+                # second run to account for files that are to be loaded whole
                 if s is not None and u not in whole_files:
                     urls2.append(u)
                     starts2.append(s)
                     ends2.append(e)
                     paths2.append(p)
+
+            # merge and fetch consolidated ranges
             new_paths, new_starts, new_ends = merge_offset_ranges(
                 list(urls2),
                 list(starts2),
                 list(ends2),
-                sort=False,
+                sort=True,
                 max_gap=self.max_gap,
                 max_block=self.max_block,
             )
             bytes_out = fs.cat_ranges(new_paths, new_starts, new_ends)
-            if len(urls2) == len(bytes_out):
-                # we didn't do any merging
-                for p, b in zip(paths2, bytes_out):
-                    out[p] = b
-            else:
-                # unbundle from merged bytes - simple approach
-                for u, s, e, p in zip(urls, starts, ends, paths):
-                    if p in out:
-                        continue  # was bytes, already handled
-                    for np, ns, ne, b in zip(
-                        new_paths, new_starts, new_ends, bytes_out
-                    ):
-                        if np == u and (ns is None or ne is None):
+
+            # unbundle from merged bytes - simple approach
+            for u, s, e, p in zip(urls, starts, ends, paths):
+                if p in out:
+                    continue  # was bytes, already handled
+                for np, ns, ne, b in zip(new_paths, new_starts, new_ends, bytes_out):
+                    if np == u and (ns is None or ne is None):
+                        if isinstance(b, Exception):
+                            out[p] = b
+                        else:
                             out[p] = b[s:e]
-                        elif np == u and s >= ns and e <= ne:
+                    elif np == u and s >= ns and e <= ne:
+                        if isinstance(b, Exception):
+                            out[p] = b
+                        else:
                             out[p] = b[s - ns : (e - ne) or None]
 
+        for k, v in out.copy().items():
+            # these were valid references, but fetch failed, so transform exc
+            if isinstance(v, Exception) and k in self.references:
+                ex = out[k]
+                new_ex = ReferenceNotReachable(k, self.references[k])
+                new_ex.__cause__ = ex
+                if on_error == "raise":
+                    raise new_ex
+                elif on_error != "omit":
+                    out[k] = new_ex
+
         if len(out) == 1 and isinstance(path, str) and "*" not in path:
             return _first(out)
         return out
 
-    def _process_dataframe(self):
-        self._process_templates(self.templates)
-
-        @lru_cache(1000)
-        def _render_jinja(url):
-            if "{{" in url:
-                if self.simple_templates:
-                    return (
-                        url.replace("{{", "{")
-                        .replace("}}", "}")
-                        .format(**self.templates)
-                    )
-
-                import jinja2
-
-                return jinja2.Template(url).render(**self.templates)
-
-            return url
-
-        if self.templates:
-            self.df["url"] = self.df["url"].map(_render_jinja)
-
     def _process_references(self, references, template_overrides=None):
-        if isinstance(references, (str, bytes)):
-            references = json.loads(references)
         vers = references.get("version", None)
         if vers is None:
             self._process_references0(references)
@@ -408,10 +425,7 @@ class ReferenceFileSystem(AsyncFileSystem):
 
     def _process_references1(self, references, template_overrides=None):
         if not self.simple_templates or self.templates:
-            try:
-                import jinja2
-            except ImportError as e:
-                raise ValueError("Reference Spec Version 1 requires jinja2") from e
+            import jinja2
         self.references = {}
         self._process_templates(references.get("templates", {}))
 
@@ -493,23 +507,14 @@ class ReferenceFileSystem(AsyncFileSystem):
 
     def _dircache_from_items(self):
         self.dircache = {"": []}
-        if self.dataframe:
-            it = self.df.iterrows()
-        else:
-            it = self.references.items()
+        it = self.references.items()
         for path, part in it:
-            if self.dataframe:
-                if part["data"]:
-                    size = len(part["data"])
-                else:
-                    size = part["size"]
+            if isinstance(part, (bytes, str)):
+                size = len(part)
+            elif len(part) == 1:
+                size = None
             else:
-                if isinstance(part, (bytes, str)):
-                    size = len(part)
-                elif len(part) == 1:
-                    size = None
-                else:
-                    _, start, size = part
+                _, start, size = part
             par = path.rsplit("/", 1)[0] if "/" in path else ""
             par0 = par
             while par0 and par0 not in self.dircache:
@@ -522,9 +527,7 @@ class ReferenceFileSystem(AsyncFileSystem):
 
             self.dircache[par].append({"name": path, "type": "file", "size": size})
 
-    def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
-        if mode != "rb":
-            raise NotImplementedError
+    def _open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
         data = self.cat_file(path)  # load whole chunk into memory
         return io.BytesIO(data)
 
@@ -570,7 +573,7 @@ class ReferenceFileSystem(AsyncFileSystem):
         if detail:
             if not self.dircache:
                 self._dircache_from_items()
-            return {k: self._ls_from_cache(k) for k in r}
+            return {k: self._ls_from_cache(k)[0] for k in r}
         else:
             return r
 
@@ -630,20 +633,324 @@ class ReferenceFileSystem(AsyncFileSystem):
             f.write(json.dumps({"version": 1, "refs": out}).encode())
 
 
+def prefix(x):
+    if "/.z" in x or "/" not in x:
+        return "metadata", x
+    return x.split("/", 1)
+
+
+def constant_prefix(x):
+    return "metadata", x
+
+
+class DFReferenceFileSystem(AbstractFileSystem):
+    """
+    (Experimental) Parquet-based Reference Filesystem
+
+    Putative replacement or adjunct to ReferenceFileSystem with
+    additional capabilities:
+    - loads from parquet for better on-disk and in-memory space
+    - optional lazy loading by key prefix (lazy=True)
+    - multiple targets for a given key (allow_multi=True), concatenated
+      together by default, of multi_func=
+    - per-chunk processing with extra parameters stored in the parquet
+      (chunk_func=)
+
+    This implementation is not (yet) multable.
+    """
+
+    def __init__(
+        self,
+        fo,
+        target_options=None,
+        remote_protocol=None,
+        remote_options=None,
+        fs=None,
+        max_gap=64_000,
+        max_block=256_000_000,
+        parquet_kwargs=None,
+        chunk_func=None,
+        allow_multi=False,
+        multi_func=b"".join,
+        prefix_func=prefix,
+        lazy=False,
+        **kwargs,
+    ):
+        self.fo = fo
+        self.target_options = target_options or {}
+        self.max_gap = max_gap
+        self.max_block = max_block
+        self.dataframes = {}
+        self.keysets = {}
+        self.url_dict = {}
+        self.template_dict = {}
+        self.prefs = None
+        self.fss = {}
+        self.dirs = None
+        self.lazy = lazy
+        self.chunk_func = chunk_func
+        self.allow_multi = allow_multi
+        self.multi_func = multi_func
+        self.prefix_func = prefix_func if lazy else constant_prefix
+        self.pkwargs = parquet_kwargs or {}
+        if fs is not None:
+            # single remote FS
+            remote_protocol = (
+                fs.protocol[0] if isinstance(fs.protocol, tuple) else fs.protocol
+            )
+            self.fss[remote_protocol] = fs
+
+        if remote_protocol and remote_protocol not in self.fss:
+            fs = filesystem(remote_protocol, **(remote_options or {}))
+            self.fss[remote_protocol] = fs
+
+        if fs:
+            self.fss[None] = fs
+        elif self.fss:
+            self.fss[None] = iter(self.fss.values()).__next__()
+        else:
+            self.fss[None] = fsspec.filesystem(
+                remote_protocol, **(remote_options or {})
+            )
+
+        super().__init__(**kwargs)
+        self._reference_part()
+
+    def _reference_part(self, part="metadata"):
+        """Load some references from parquet
+
+        If lazy is False, this is called exactly once per instance
+
+        If lazy is true, selecting a path will determine the name of
+        the target parquet file, and the resultant columns will be
+        cached so the file need not be read again
+        """
+        import fastparquet
+
+        if part != "metadata" and part not in self.dirs:
+            raise FileNotFoundError(f"prefix {part}")
+        if part not in self.dataframes:
+            url = f"{self.fo}/{part}.parq" if self.lazy else self.fo
+            fs, path = fsspec.core.url_to_fs(url, **self.target_options)
+            pf = fastparquet.ParquetFile(path, fs=fs)
+            self.template_dict[part] = pf.key_value_metadata
+            df = pf.to_pandas()
+            thispart = {}
+            for k in df:
+                if df[k].dtype == "category" and k == "path":
+                    self.url_dict[part] = df[k].cat.categories.values
+                    thispart[k] = df[k].cat.codes.values
+                else:
+                    thispart[k] = df[k].values
+            self.dataframes[part] = thispart
+            if self.allow_multi is False:
+                self.keysets[part] = {
+                    k: i for (i, k) in enumerate(self.dataframes[part]["key"])
+                }
+            else:
+                self.keysets[part] = {}
+                for i, k in enumerate(self.dataframes[part]["key"]):
+                    self.keysets[part].setdefault(k, []).append(i)
+            if part == "metadata":
+                self.dirs = {
+                    k.rsplit("/", 1)[0]
+                    for k in self.dataframes[part]["key"]
+                    if "/" in k
+                }
+                self.prefs = (
+                    ast.literal_eval(pf.key_value_metadata["prefs"])
+                    if "prefs" in pf.key_value_metadata
+                    else set()
+                )
+
+        return self.dataframes[part]
+
+    def isdir(self, path):
+        return path in self.dirs
+
+    def cat_file(self, path, start=None, end=None, **kwargs):
+        return self.cat_ranges([path], [start], [end])[0]
+
+    def cat(self, path, recursive=False, on_error="return", **kwargs):
+        paths = self.expand_path(path, recursive=recursive)
+        paths1 = [p for p in paths if not self.isdir(p)]
+        result = {
+            p: data
+            for p, data in zip(
+                paths1, self.cat_ranges(paths1, on_error=on_error, **kwargs)
+            )
+        }
+        if len(paths1) == 1 and recursive is False and "*" not in path:
+            # same as cat_file
+            return list(result.values())[0]
+        return result
+
+    def cat_ranges(self, paths, starts=None, ends=None, on_error="return", **kwargs):
+        out = []  # eventual output; initially each key contains raw bytes or None
+        proto_dict = {}  # mapping of protocol to lists of URL/start/end to fetch
+        assign_dict = {}  # how to assign the results of cat_ranges to output
+        if starts is None:
+            starts = [None] * len(paths)
+        if ends is None:
+            ends = [None] * len(paths)
+        for p, s, e in zip(paths, starts, ends):
+            thislist = []
+            out.append(thislist)
+            if self.lazy:
+                pref, p0 = self.prefix_func(p)
+                if pref in self.prefs:
+                    # reference already inlined in metadata file
+                    pref = "metadata"
+                else:
+                    # new key in the target pref file
+                    p = p0
+            else:
+                # everything is in the same file
+                pref = "metadata"
+            self._reference_part(pref)
+            inds = self.keysets[pref][p]
+            if isinstance(inds, int):
+                inds = [inds]
+            for i in inds:
+                if x := self.dataframes[pref]["raw"][i]:
+                    thislist.append(x)
+                else:
+                    # infer path - cache this?
+                    path = self.dataframes[pref]["path"][i]
+                    if pref in self.url_dict:
+                        # dict-encoded columns; actually, numpy can do
+                        # many of these at once with int fancy indexing
+                        path = self.url_dict[pref][path]
+                    # apply template: common prefix
+                    path = path.format(**self.template_dict[pref])
+
+                    prot, _ = split_protocol(path)
+                    proto_dict.setdefault(prot, [[], [], []])
+                    proto_dict[prot][0].append(path)
+                    if s is None or s >= 0:
+                        proto_dict[prot][1].append(
+                            self.dataframes[pref]["offset"][i] or s
+                        )
+                    else:
+                        # range is from end of file, which we do not know
+                        # the size of, so this can only work if there is no
+                        # merging
+                        proto_dict[prot][1].append(s)
+
+                    if e is None or e >= 0:
+                        proto_dict[prot][2].append(
+                            self.dataframes[pref]["offset"][i]
+                            + self.dataframes[pref]["size"][i]
+                            or e
+                        )
+                    else:
+                        # range is from end of file, which we do not know
+                        # the size of, so this can only work if there is no
+                        # merging
+                        proto_dict[prot][1].append(e)
+                    thislist.append(None)
+                    assign_dict.setdefault(prot, []).append(
+                        (thislist, len(thislist) - 1)
+                    )
+
+        for proto, (urls2, starts2, ends2) in proto_dict.items():
+            fs = self.fss[proto]
+
+            new_paths, new_starts, new_ends = merge_offset_ranges(
+                list(urls2),
+                list(starts2),
+                list(ends2),
+                sort=True,
+                max_gap=self.max_gap,
+                max_block=self.max_block,
+            )
+            bytes_out = fs.cat_ranges(new_paths, new_starts, new_ends)
+            if len(urls2) == len(bytes_out):
+                # we didn't do any merging
+                for (l, i), d in zip(assign_dict[proto], bytes_out):
+                    l[i] = d
+            else:
+                # unbundle from merged bytes - simple approach
+                for u, s, e, (l, i) in zip(urls2, starts2, ends2, assign_dict[proto]):
+                    if p in out:
+                        continue  # was bytes, already handled
+                    for np, ns, ne, b in zip(
+                        new_paths, new_starts, new_ends, bytes_out
+                    ):
+                        if np == u and (ns is None or ne is None):
+                            l[i] = b[s:e]
+                        elif np == u and s >= ns and e <= ne:
+                            l[i] = b[s - ns : (e - ne) or None]
+
+        out = [self.multi_func(part) for part in out]
+        return out
+
+    def find(self, path, detail=False, withdirs=False, **kwargs):
+        path = self._strip_protocol(path)
+        if path in self.dirs:
+            path = path + "/"
+        pref, p = self.prefix_func(path)
+        dirs = (
+            [
+                {"name": d, "size": 0, "type": "directory"}
+                for d in self.dirs
+                if d.startswith(path)
+            ]
+            if withdirs
+            else []
+        )
+        if pref in self.prefs:
+            pref = "metadata"
+        df = self._reference_part(pref)
+        files = [
+            {"name": k, "type": "file", "size": _size(self.dataframes["metadata"], i)}
+            for k, i in self.keysets["metadata"].items()
+            if k.startswith(path)
+        ]
+        if self.lazy and pref != "metadata":
+            files.extend(
+                [
+                    {"name": f"{pref}/{k}", "type": "file", "size": _size(df, i)}
+                    for k, i in self.keysets[pref].items()
+                    if k.startswith(p)
+                ]
+            )
+        if detail:
+            return dirs + files
+        return [k["name"] for k in dirs + files]
+
+    def ls(self, path, detail=True, **kwargs):
+        path = self._strip_protocol(path)
+        allfiles = self.find(path, detail=True, withdirs=True)
+        isdir = path in self.dirs
+        subdfiles = [
+            p for p in allfiles if p["name"].count("/") == path.count("/") + isdir
+        ]
+        if detail:
+            return subdfiles
+        return [p["name"] for p in subdfiles]
+
+    def info(self, path, **kwargs):
+        path = self._strip_protocol(path)
+
+        if path in self.dirs:
+            return {"name": path, "type": "directory", "Size": 0}
+        return self.ls(path, detail=True)[0]
+
+
+def _size(df, i):
+    if isinstance(i, int):
+        return len(df["raw"][i]) if df["raw"][i] else df["size"][i]
+    return sum(len(df["raw"][_]) if df["raw"][_] else df["size"][_] for _ in i)
+
+
 def _unmodel_hdf5(references):
     """Special JSON format from HDF5 prototype"""
     # see https://gist.github.com/ajelenak/80354a95b449cedea5cca508004f97a9
-    import re
-
     ref = {}
     for key, value in references["metadata"].items():
         if key.endswith(".zchunkstore"):
             source = value.pop("source")["uri"]
-            match = re.findall(r"https://([^.]+)\.s3\.amazonaws\.com", source)
-            if match:
-                source = source.replace(
-                    f"https://{match[0]}.s3.amazonaws.com", match[0]
-                )
             for k, v in value.items():
                 ref[k] = (source, v["offset"], v["offset"] + v["size"])
         else:
