@@ -1,43 +1,15 @@
 import importlib
+import types
+import warnings
 
 __all__ = ["registry", "get_filesystem_class", "default"]
 
-# mapping protocol: implementation class object
-_registry = {}  # internal, mutable
+# internal, mutable
+_registry = {}
 
-
-class ReadOnlyError(TypeError):
-    pass
-
-
-class ReadOnlyRegistry(dict):
-    """Dict-like registry, but immutable
-
-    Maps backend name to implementation class
-
-    To add backend implementations, use ``register_implementation``
-    """
-
-    def __init__(self, target):
-        self.target = target
-
-    def __getitem__(self, item):
-        return self.target[item]
-
-    def __delitem__(self, key):
-        raise ReadOnlyError
-
-    def __setitem__(self, key, value):
-        raise ReadOnlyError
-
-    def clear(self):
-        raise ReadOnlyError
-
-    def __contains__(self, item):
-        return item in self.target
-
-    def __iter__(self):
-        yield from self.target
+# external, immutable
+registry = types.MappingProxyType(_registry)
+default = "file"
 
 
 def register_implementation(name, cls, clobber=True, errtxt=None):
@@ -79,9 +51,6 @@ def register_implementation(name, cls, clobber=True, errtxt=None):
         _registry[name] = cls
 
 
-registry = ReadOnlyRegistry(_registry)
-default = "file"
-
 # protocols mapped to the class which implements them. This dict can
 # updated with register_implementation
 known_implementations = {
@@ -126,7 +95,7 @@ known_implementations = {
     },
     "ftp": {"class": "fsspec.implementations.ftp.FTPFileSystem"},
     "hdfs": {
-        "class": "fsspec.implementations.hdfs.PyArrowHDFS",
+        "class": "fsspec.implementations.arrow.HadoopFileSystem",
         "err": "pyarrow and local java libraries required for HDFS",
     },
     "arrow_hdfs": {
@@ -144,6 +113,10 @@ known_implementations = {
         "class": "ocifs.OCIFileSystem",
         "err": "Install ocifs to access OCI Object Storage",
     },
+    "asynclocal": {
+        "class": "morefs.asyn_local.AsyncLocalFileSystem",
+        "err": "Install 'morefs[asynclocalfs]' to use AsyncLocalFileSystem",
+    },
     "adl": {
         "class": "adlfs.AzureDatalakeFileSystem",
         "err": "Install adlfs to access Azure Datalake Gen1",
@@ -202,12 +175,17 @@ known_implementations = {
         "class": "webdav4.fsspec.WebdavFileSystem",
         "err": "Install webdav4 to access WebDAV",
     },
+    "dvc": {
+        "class": "dvc.api.DVCFileSystem",
+        "err": "Install dvc to access DVCFileSystem",
+    },
     "root": {
         "class": "fsspec_xrootd.XRootDFileSystem",
         "err": "Install fsspec-xrootd to access xrootd storage system."
         + " Note: 'root' is the protocol name for xrootd storage systems,"
-        + " not refering to root directories",
+        + " not referring to root directories",
     },
+    "dir": {"class": "fsspec.implementations.dirfs.DirFileSystem"},
 }
 
 
@@ -264,6 +242,13 @@ def filesystem(protocol, **storage_options):
     ``storage_options`` are specific to the protocol being chosen, and are
     passed directly to the class.
     """
+    if protocol == "arrow_hdfs":
+        warnings.warn(
+            "The 'arrow_hdfs' protocol has been deprecated and will be "
+            "removed in the future. Specify it as 'hdfs'.",
+            DeprecationWarning,
+        )
+
     cls = get_filesystem_class(protocol)
     return cls(**storage_options)
 
