@@ -848,25 +848,26 @@ class AbstractFileSystem(metaclass=_Cached):
         """Copy single remote file to local"""
         if isfilelike(lpath):
             outfile = lpath
-        else:
-            if self.isdir(rpath):
-                os.makedirs(lpath, exist_ok=True)
-                return None
+        elif self.isdir(rpath):
+            os.makedirs(lpath, exist_ok=True)
+            return None
 
+        with self.open(rpath, "rb", **kwargs) as f1:
             if outfile is None:
                 outfile = open(lpath, "wb")
 
-        with self.open(rpath, "rb", **kwargs) as f1:
-            callback.set_size(getattr(f1, "size", None))
-            data = True
-            while data:
-                data = f1.read(self.blocksize)
-                segment_len = outfile.write(data)
-                if segment_len is None:
-                    segment_len = len(data)
-                callback.relative_update(segment_len)
-        if not isfilelike(lpath):
-            outfile.close()
+            try:
+                callback.set_size(getattr(f1, "size", None))
+                data = True
+                while data:
+                    data = f1.read(self.blocksize)
+                    segment_len = outfile.write(data)
+                    if segment_len is None:
+                        segment_len = len(data)
+                    callback.relative_update(segment_len)
+            finally:
+                if not isfilelike(lpath):
+                    outfile.close()
 
     def get(self, rpath, lpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) to local.
@@ -878,12 +879,20 @@ class AbstractFileSystem(metaclass=_Cached):
 
         Calls get_file for each source.
         """
-        from .implementations.local import make_path_posix
+        from .implementations.local import LocalFileSystem, make_path_posix
 
         if isinstance(lpath, str):
             lpath = make_path_posix(lpath)
         rpaths = self.expand_path(rpath, recursive=recursive)
-        lpaths = other_paths(rpaths, lpath)
+        isdir = isinstance(lpath, str) and (
+            lpath.endswith("/") or LocalFileSystem().isdir(lpath)
+        )
+        lpaths = other_paths(
+            rpaths,
+            lpath,
+            exists=isdir and isinstance(rpath, str) and not rpath.endswith("/"),
+            is_dir=isdir,
+        )
 
         callback.set_size(len(lpaths))
         for lpath, rpath in callback.wrap(zip(lpaths, rpaths)):
@@ -930,12 +939,12 @@ class AbstractFileSystem(metaclass=_Cached):
             lpath = make_path_posix(lpath)
         fs = LocalFileSystem()
         lpaths = fs.expand_path(lpath, recursive=recursive)
-        isdir = isinstance(rpath, str) and self.isdir(rpath)
+        isdir = isinstance(rpath, str) and (rpath.endswith("/") or self.isdir(rpath))
         rpaths = other_paths(
             lpaths,
             rpath,
-            exists=isdir,
-            is_dir=isdir and len(lpaths) == 1,
+            exists=isdir and isinstance(lpath, str) and not lpath.endswith("/"),
+            is_dir=isdir,
         )
 
         callback.set_size(len(rpaths))
@@ -971,7 +980,14 @@ class AbstractFileSystem(metaclass=_Cached):
             on_error = "raise"
 
         paths = self.expand_path(path1, recursive=recursive)
-        path2 = other_paths(paths, path2)
+        isdir = isinstance(path2, str) and (path2.endswith("/") or self.isdir(path2))
+        path2 = other_paths(
+            paths,
+            path2,
+            exists=isdir and isinstance(path1, str) and not path1.endswith("/"),
+            is_dir=isdir,
+        )
+
         for p1, p2 in zip(paths, path2):
             try:
                 self.cp_file(p1, p2, **kwargs)
@@ -1191,7 +1207,7 @@ class AbstractFileSystem(metaclass=_Cached):
         offset: int
             Byte offset to start read
         length: int
-            Number of bytes to read
+            Number of bytes to read. If None, read to end.
         delimiter: bytes (optional)
             Ensure reading starts and stops at delimiter bytestring
 
