@@ -7,6 +7,7 @@ import sys
 from contextlib import contextmanager
 from functools import partial
 from hashlib import md5
+from importlib.metadata import version
 from urllib.parse import urlsplit
 
 DEFAULT_BLOCK_SIZE = 5 * 2**20
@@ -365,16 +366,10 @@ def other_paths(paths, path2, is_dir=None, exists=False):
     if isinstance(path2, str):
         is_dir = is_dir or path2.endswith("/")
         path2 = path2.rstrip("/")
-        if len(paths) > 1:
-            cp = common_prefix(paths)
-            if exists:
-                cp = cp.rsplit("/", 1)[0]
-            path2 = [p.replace(cp, path2, 1) for p in paths]
-        else:
-            if is_dir:
-                path2 = [path2.rstrip("/") + "/" + paths[0].rsplit("/")[-1]]
-            else:
-                path2 = [path2]
+        cp = common_prefix(paths)
+        if exists:
+            cp = cp.rsplit("/", 1)[0]
+        path2 = [p.replace(cp, path2, 1) for p in paths]
     else:
         assert len(paths) == len(path2)
     return path2
@@ -422,20 +417,10 @@ def get_package_version_without_import(name):
         mod = sys.modules[name]
         if hasattr(mod, "__version__"):
             return mod.__version__
-    if sys.version_info >= (3, 8):
-        try:
-            import importlib.metadata
-
-            return importlib.metadata.distribution(name).version
-        except:  # noqa: E722
-            pass
-    else:
-        try:
-            import importlib_metadata
-
-            return importlib_metadata.distribution(name).version
-        except:  # noqa: E722
-            pass
+    try:
+        return version(name)
+    except:  # noqa: E722
+        pass
     try:
         import importlib
 
@@ -496,7 +481,6 @@ def merge_offset_ranges(paths, starts, ends, max_gap=0, max_block=None, sort=Tru
     order. If the user can guarantee that the inputs are already
     sorted, passing `sort=False` will skip the re-ordering.
     """
-
     # Check input
     if not isinstance(paths, list):
         raise TypeError
@@ -511,9 +495,17 @@ def merge_offset_ranges(paths, starts, ends, max_gap=0, max_block=None, sort=Tru
     if len(starts) <= 1:
         return paths, starts, ends
 
+    starts = [s or 0 for s in starts]
     # Sort by paths and then ranges if `sort=True`
     if sort:
-        paths, starts, ends = [list(v) for v in zip(*sorted(zip(paths, starts, ends)))]
+        paths, starts, ends = [
+            list(v)
+            for v in zip(
+                *sorted(
+                    zip(paths, starts, ends),
+                )
+            )
+        ]
 
     if paths:
         # Loop through the coupled `paths`, `starts`, and
@@ -522,7 +514,9 @@ def merge_offset_ranges(paths, starts, ends, max_gap=0, max_block=None, sort=Tru
         new_starts = starts[:1]
         new_ends = ends[:1]
         for i in range(1, len(paths)):
-            if (
+            if paths[i] == paths[i - 1] and new_ends[-1] is None:
+                continue
+            elif (
                 paths[i] != paths[i - 1]
                 or ((starts[i] - new_ends[-1]) > max_gap)
                 or ((max_block is not None and (ends[i] - new_starts[-1]) > max_block))
